import sys
import os
import re
import string
import types
import time

import libxml2

# TODO: move parts of it to a config file?
# Syntax of "ignore": regexp of files to ignore. By default: "NOT-YET"
ignore_patterns = "./ignore"
# Syntax of "pinned": one line, files separated by whitespace, full
# name, with directory, e.g. entries/dangers-sur-internet.entry_xml
# WARNING: because of a bug, do not use "pinned" if the file would be
# first, anyway. You risk duplication.
pinned_files_names = "./pinned"
#
start = os.environ["HOME"] + os.sep + "/Blog"
# All my sections. TODO: track places in the code where this list is hardwired
dirs = ["RFC", "entries", "ficheslec"]
# Canonical URL
url_prefix = "http://www.bortzmeyer.org/"

def by_date(item):
    return time.strptime(date_of(item[1]), "%Y-%m-%d")

def by_rfcnum(item):
    return rfcnum_of(item[1])

def file2DOM(filename):
    doc = libxml2.parseFile(filename)
    return doc

def has_summary(filename):
    summary_file = re.sub(r"_xml$", "_summary.html", filename)
    return os.path.exists(summary_file)
    
def htmlfilename_of(filename, final=False, summary=False):
    if not final:
        if not summary:
            return re.sub(r"_xml$", ".html", filename)
        else:
            summary_file = re.sub(r"_xml$", "_summary.html", filename)
            if os.path.exists(summary_file):
                return summary_file
            else:
                return re.sub(r"_xml$", ".html", filename)
    else:
        if re.search(r"\.entry_xml$", filename):
            return os.path.basename(re.sub(r"\.entry_xml$", ".html", filename))
        elif re.search(r"\.rfc_xml$", filename):
            return os.path.basename(re.sub(r"\.rfc_xml$", ".html", filename))
        elif re.search(r"\.fiche_xml$", filename):
            return os.path.basename(re.sub(r"\.fiche_xml$", ".html", filename))
        else:
            raise Exception("Unknown type for file name %s" % filename)
    
def html_of(filename, summary=False):
    htmlfilename = htmlfilename_of(filename, summary=summary)
    return open(htmlfilename).read()

def sanitize(xpath_expr):
    """ Turns a XPath expression into a legal Unix file name.
        TODO: a better replacement, to limit the risk of collisions. """
    return re.sub(r"/|@|\*|\.|\(|\)", "_", xpath_expr)

def xpath_eval(dom, xpath_expr, mandatory=True):
    match = dom.xpathNewContext().xpathEval(xpath_expr)
    if match:
        return match[0].content
    else:
        if mandatory:
            raise Exception("No \"%s\" in the DOM tree %s" % (xpath_expr, dom.name))
        else:
            return None

def xpath_from(dom, xpath_expr, use_cache=False, mandatory=True):
    """ use_cache is False by default because, in most cases, it gives you nothing. """
    if use_cache:
        filecachename = re.sub(r"\.[^\.]+$", ".%s.xpath-cache" % sanitize(xpath_expr),
                               dom.name)
        if os.path.exists (filecachename): # We do not check its creation time (to be sure
            # it is newer than filename). We could rely on "make" to do so.
            result = open(filecachename).read()
        else:
            result = xpath_eval(dom, xpath_expr, mandatory)
            cache = open(filecachename, "w")
            cache.write("%s" % result)
            cache.close()
    else:
        result = xpath_eval(dom, xpath_expr, mandatory)
    return result

def date_of(dom):
    return xpath_from(dom, "/*/date/text()")

def update_date_of(dom):
    try:
        return xpath_from(dom, "/*/update_date/text()")
    except: # Unfortunately, libxml2 returns a generic exception if
            # the Xpath expression founds nothing :-(
        return xpath_from(dom, "/*/date/text()")

def lang_of(dom):
    try:
        return xpath_from(dom, "/*/@xml:lang")
    except: # Unfortunately, libxml2 returns a generic exception if
            # the Xpath expression founds nothing :-(
        return None

def title_of(dom):
    title = xpath_from(dom, "/*/@title")
    match = xpath_from(dom, "/rfcdesc/@num", mandatory=False)
    if match:
        # It is a RFC
        # Sanity check
        number = int(match)
        title = "RFC %i: %s" % (number, title)
    return title

def text_content_of(element):
    result = ""
    # Special elements
    if element.tag == "rfcdesc":
        if "wg" in element.attrib:
            result += " (groupe de travail %s)" % element.attrib["wg"]
    if element.tag == "rfc":
        result = result + "RFC " + element.attrib["num"] + " "
    elif element.tag == "link":
        if "url" in element.attrib:
            url = element.attrib["url"]
        else:
            url = url_prefix + element.attrib["local"] + ".html"
        if element.text is None or element.text == "":
            result = result + "<" + url + ">"
        else:
            result = result + element.text + " <" + url + ">"
    else:  # General case
        if element.text is not None:
            result = result + " " + element.text
        for sub in element:
            result = result + " " + text_content_of(sub)
    if element.tail is not None:
        result = result + " " + element.tail
    return result

def wikipedia_of(element, parent_lang=None):
    result = []
    if element.tag != "wikipedia":
        if "{http://www.w3.org/XML/1998/namespace}lang" in element.attrib:
            parent_lang = element.attrib["{http://www.w3.org/XML/1998/namespace}lang"]
        for sub in element:
            result = result + wikipedia_of(sub, parent_lang)
    else:
        if "{http://www.w3.org/XML/1998/namespace}lang" in element.attrib:
            lang = element.attrib["{http://www.w3.org/XML/1998/namespace}lang"]
        else:
            if parent_lang is not None:
                lang = parent_lang
            else:
                lang = 'fr'
        if "name" in element.attrib:
            name = element.attrib["name"]
        else:
            name = element.text
        if len(name) == 0:
            raise Exception("Empty Wikipedia name")
        elif len(name) == 1:
            name = name.upper()
        else:
            name = name[0].upper() + name[1:]
        name = re.sub(r"\n *", " ", name)
        name = re.sub(r" +", " ", name)
        result = result + [{'lang': lang, 'name': name}]
    return result

def link_of(element):
    result = []
    if element.tag != "link":
        for sub in element:
            result = result + link_of(sub)
    else:
        if "local" in element.attrib:
            pass
        elif "url" in element.attrib:
            url = element.attrib["url"]
            if re.search(r"^[a-z]+:", url):
                result = result + [url,]
            else:
                # Pseudo-remote link, for instance to /search
                pass
        else:
            raise Exception("Element <link> without attribute url or local")
    return result

def most_recent_change(dom):
    changed = update_date_of(dom)
    if changed is None:
        changed = date_of(dom)
    return changed

def rfcnum_of(dom):
    return int(xpath_from(dom, "/rfcdesc/@num"))
    # TODO: suitable error message if there is no @num
    
def list_xml(what):
    """ Return a list of doubles. The first item of the double is the
    file name, the second is the DOM tree.
    The argument what is either a string, the directory name, or an
    array, the file names."""
    results = []
    s = open(ignore_patterns, "r").readline()
    ignore_text = s.strip()
    ignore_text = ignore_text + "|PLACEHOLDERS"
    ignore = re.compile(ignore_text)
    if type(what) in (str,):
        directory = True
        mydir = what
        files = os.listdir(mydir)
    else:
        directory = False
        files = what
    for file in files:
        datafile = re.search(r"\.[A-Za-z0-9]+_xml$", file)
        if ignore.search(file):
            continue
        if not datafile:
            continue
        if directory:
            file_fullname = "%s/%s" % (mydir, file)
        else:
            file_fullname = file
        root = file2DOM(file_fullname)
        results.append((file_fullname, root))
    return results

def entries(thedir=".", max=None, year=None, serie=None, sort_by=by_date):
    """ Returns a list of doubles [file_name, date] """
    if year == "":
        year = None
    if year is not None:
        year = int(year)
    if serie == "":
        serie = None
    if serie is not None:
        serie = int(serie)
    if type(thedir) in (str,):
        dirs = [thedir,]
    else:
        dirs = thedir
    files = []
    results = []
    s = open(pinned_files_names, "r").readline()
    pinned_files_all = re.split(r"[ ;,]", s.strip())
    pinned_files = []
    for file in pinned_files_all:
        for dir in dirs:
            if re.search(r"^(\./)?%s/" % dir, file):
                pinned_files.append(file)
    for mydir in dirs:
        for myfile in list_xml(mydir):
            if myfile[0] not in pinned_files:
                files.append(myfile)
    files.sort(key=sort_by, reverse=True)
    files = list_xml(pinned_files) + files
    if max is not None:
        num = 0
    for myfile in files:
        date = date_of(myfile[1])
        if year is not None:
            yearfile = int(date.split('-')[0])
            if year != yearfile:
                continue
        if serie is not None:
            num = rfcnum_of(myfile[1])
            try:
                myserie = num - (num % 1000)
            except:
                sys.stderr.write("Error while handling RFC %s\n" % myfile[0])
                raise
            if myserie != serie:
                continue
        results.append([myfile[0], date])
        if max is not None:
            num = num + 1
            if num >= max:
                break
    if len(results) == 0:
        if year is not None:
            raise Exception("No entries in %s for %i" % (thedir, year))
        elif serie is not None:
            raise Exception("No RFC in serie %s" % serie)
        else:
            raise Exception("No entries in %s" % thedir)
    return results

def available_years(thedir=".", max=None):
    if type(thedir) in (str,):
        dirs = [thedir,]
    else:
        dirs = thedir
    files = []
    results = {}
    for mydir in dirs:
        for myfile in list_xml(mydir):
            files.append(myfile)
    for myfile in files:
        year = date_of(myfile[1]).split('-')[0]
        results[year] = True
    results = list(results.keys())
    results.sort()
    results.reverse()
    if max is not None:
        if max >= len(results):
            end = None
        else:
            end = max
        return results[:end]
    else:
        return results

def available_rfc_series(thedir=".", max=None):
    if type(thedir) in (str,):
        dirs = [thedir,]
    else:
        dirs = thedir
    files = []
    results = {}
    for mydir in dirs:
        for myfile in list_xml(mydir):
            files.append(myfile)
    for myfile in files:
        num = rfcnum_of(myfile[1])
        serie = num - (num % 1000) # Zero if the RFC < 1000
        results[str(serie)] = True
    results = list(results.keys())
    results.sort()
    if max is not None:
        if max >= len(results):
            start = 0
        else:
            start = len(results) - max
        return results[start:]
    else:
        return results

def all_files(base, directory):
    """ Returns an array of 2-tuples (file, complete_name) """
    result = []
    if directory is not None:
        if not os.path.isdir(directory):
            print("%s does not exist or is not a directory" % directory, file=sys.stderr)
            sys.exit(1)
        for dir in dirs:
            for file in os.listdir(directory + os.sep + dir):
                filename = start + os.sep + dir + os.sep + file
                if re.search(r"\.(entry|rfc|fiche)\_xml$", file) and \
                       not re.search(r"-(NOT-YET|PLACEHOLDERS)\.", file):
                    result.append((file, filename))
    elif base is not None:
        if not os.path.isdir(base):
            print("%s does not exist or is not a directory" % base, file=sys.stderr)
            sys.exit(1)
        for file in os.listdir(base):
            filename = base + os.sep + file
            if re.search(r"\.xml$", file) and \
                   not re.search(r"(schemas|sitemap|-(NOT-YET|PLACEHOLDERS))\.", file):
                result.append((file, filename))
    else:
        raise Exception("Internal error, all_files called with both base and directory None")
    return result
 
if __name__ == '__main__':
    print("Available RFC series: %s" % available_rfc_series("RFC", max=7))
    print("")
    print("Available years: %s" % available_years(["entries", "RFC", "ficheslec"], max=7))
    print("Available years (max 2): %s" % available_years("entries", max=2))
    print("")
    print("The first thirty entries: %s" % entries("entries", max=30))
    print("")
    print("All the entries of 2004: %s" % entries("entries", year=2004, max=3))
