#!/usr/bin/env python3

import os
import glob
import random
import sys
import json
import re
import netaddr
import socket
import collections
import time
import datetime
import urllib.parse

# https://github.com/funny-falcon/flunn (for CBOR)
import flunn

# https://pyyaml.org/
import yaml

# https://requests.readthedocs.io/en/master/
import requests

import LeakyBucket

calls = 0
pid = os.getpid()
generator = random.Random()

# For rate-limiting
buckets = collections.defaultdict(LeakyBucket.LeakyBucket)
whitelist = [netaddr.IPAddress("127.0.0.1"), netaddr.IPAddress("::1")]

path = '/var/www/www.bortzmeyer.org'
prefix = 'https://www.bortzmeyer.org'
configfile = "/etc/wsgis-bortzmeyer.yaml"
sockfile = '/home/whichasn/run/whichasn.socket'

try:
    configh = open(configfile)
    configdata = configh.read()
    # May raise yaml.parser.ParserError
    config = yaml.safe_load(configdata)
except FileNotFoundError:
    config = None
    
def env(start_response, environ):
    status = '200 OK'
    output = ""
    for key in list(environ.keys()):
        output += "%s: %s\n" % (key, environ[key])
    response_headers = [('Content-Type', 'text/plain'),
                        ('Content-Length', str(len(output)))]
    start_response(status, response_headers)
    
    return [output.encode()]

def addresses_of(addr):
    try:
        (primary, secondaries, garbage) = socket.gethostbyaddr(addr)
    except socket.herror: # Typically a NXDOMAIN but can be any error:
                          # same thing for us, we return an empty list
        return None                  
    if len(secondaries) == 0:
        result = primary
    else:
        result = [primary,]
        for name in secondaries:
            result.append(name)
    return result

def http(start_response, environ):
    status = '200 OK'
    answer = {'origin': environ['REMOTE_ADDR'],
              'method': environ['REQUEST_METHOD'],
              'headers': {}}
    names = addresses_of(environ['REMOTE_ADDR'])
    if names is not None:
        answer['origin_name'] = names
    for key in list(environ.keys()):
        if key.startswith('HTTP_'):
            skey = key[len('HTTP_'):]
            answer['headers'][skey] = environ[key]
    output = json.dumps(answer, indent=2)
    response_headers = [('Content-Type', 'application/json'),
                        ('Content-Length', str(len(output)))]
    start_response(status, response_headers) 
    return [output.encode()]

def counter(start_response, client, wsgi_version):
    status = '200 OK'
    
    output = """
    <html><head><title>WSGI works</title></head>
    <body>
    <h1>WSGI works</h1>
    <p>You are the %i th visitor for this WSGI daemon (Process ID %i).
    Reload the page to see the changes.</p>
    <p>You arrived from machine %s.</p>
    </body>
    </html>""" % (calls, pid, client)
    
    response_headers = [('Content-type', 'text/html'),
                        ('Content-Length', str(len(output))),
                        ('Server', "Apache wsgi/%s Python/%s.%s.%s" % \
                         (wsgi_version,
                          sys.version[0],sys.version[2],sys.version[4]))]
    # TODO: an Expires field of now?
    # TODO: a Server field. The one set by the script seems ignored.
    start_response(status, response_headers)

    return [output.encode()]

def date_in_cbor(start_response, client):                        
    status = '200 OK'

    now = int(time.time())
    now_tai = int(time.clock_gettime(time.CLOCK_TAI))
    now_ns = time.time_ns()
    uptime = int(time.clock_gettime(time.CLOCK_BOOTTIME))
    # RFC 8949
    as_text = flunn.Tagging(0, time.strftime("%Y-%m-%dT%H:%M:%SZ", time.gmtime(now)))
    as_num_seconds = flunn.Tagging(1, now)
    # RFC 8943
    as_num_days = flunn.Tagging(100, (datetime.datetime.utcfromtimestamp(now) - datetime.datetime(1970,1,1)).days)
    as_dateonly = flunn.Tagging(1004, time.strftime("%Y-%m-%d", time.gmtime(now)))
    # RFC 9581
    ext_time = flunn.Tagging(1001,
                             {1: now,
                              # Time scale UTC. 
                              -1: 0,
                              # Nanoseconds to add
                              -9: now_ns - now*(10**9)
                              # How to describe the characteristics of
                              # the clock (VM + NTP)?
                              })
    tai_time = flunn.Tagging(1001,
                             {1: now_tai,
                              # Time scale TAI. WARNING: it requires
                              # the kernel to be properly configured
                              # (see
                              # <https://www.bortzmeyer.org/tai-on-debian.html>
                              # for Debian). Otherwise, you won(t get
                              # an error, you will get a wrong value.
                              -1: 1
                              })
    duration = flunn.Tagging(1002,
                             {1: uptime
                              })
    output = flunn.dumps(["Current date in CBOR, done with Python %s and the flunn library <%s>" % \
                          (sys.version, "https://github.com/funny-falcon/flunn"),
                          as_text, as_num_seconds, as_num_days, as_dateonly,
                          ext_time, tai_time, "Duration since boot: ", duration])

    response_headers = [('Content-type', 'application/cbor'),                    
                        ('Content-Length', str(len(output)))] 
    # TODO: an Expires field of now?
    
    start_response(status, response_headers)                         
    return [output]                        

# RFC 9164
def addresses_in_cbor(start_response, client):
    status = "200 OK"
    baddr = netaddr.IPAddress(client)
    if config is not None and "whichasn" in config:
        received = requests.get("%s/%s" % (config["whichasn"], client)).text
    else:
        sock = socket.socket(socket.AF_UNIX, socket.SOCK_STREAM)
        try:
            sock.connect(sockfile)
            sock.sendall(bytes(client + "\n", "utf-8"))
            received = str(sock.recv(4096), "utf-8")
        finally:
            sock.close()
    (prefix, asn) = received.split(" ", maxsplit=1)
    (network, length) = prefix.split("/")
    length = int(length)
    tnetwork = netaddr.IPAddress(network).packed
    bnetwork = b""
    for byte in tnetwork:
        if byte == 0: # Section 4.2 of the RFC
            break
        bnetwork += byte.to_bytes(1, "big")
    if baddr.version == 4:
        remote = flunn.Tagging(52, baddr.packed)
        remote_prefix = flunn.Tagging(52, [length, bnetwork])
    elif baddr.version == 6:
        remote = flunn.Tagging(54, baddr.packed)
        remote_prefix = flunn.Tagging(54, [length, bnetwork])
    else:
        raise Exception("Invalid IP address family %s" % baddr.version)
    output = flunn.dumps(["Your IP address in CBOR, done with Python %s and the flunn library <%s>" % \
                          (sys.version, "https://github.com/funny-falcon/flunn"),
                          remote,
                          remote_prefix,
    ])
    response_headers = [("Content-type", "application/cbor"),
                        ("Content-Length", str(len(output)))]
    start_response(status, response_headers)
    return [output]

def randomfile(start_response, environ):
    global generator
    files = glob.glob("%s/*.xml" % path)
    file = re.sub(path, "", re.sub(r"\.xml$", r".html", generator.choice(files)))
    url = "%s%s" % (prefix, file)
    status = '307 Redirect'
    output = """
<html><head><title>Redirect at random</title></head>
    <body>
    <h1>Redirect at random</h1>
    <p>Go <a href="%s">to this article</a>.</p>
    </body>
</html>""" % url
    response_headers = [('Location', url),
                        ('Content-type', 'text/html'),
                        ('Content-Length', str(len(output)))]
    start_response(status, response_headers) 
    return [output.encode()]    
    
def slow(start_response, environ):
    form = urllib.parse.parse_qs(environ['QUERY_STRING'])
    delay = 5
    if "delay" in form:
        delay = int(form["delay"][0])
    time.sleep(delay)
    output = "Waited %i seconds" % delay
    status = '200 OK'
    response_headers = []
    start_response(status, response_headers)
    return [output.encode()]
    
def payme(start_response, environ):
    form = urllib.parse.parse_qs(environ['QUERY_STRING'])
    response_headers = []
    amount = 0
    if "pay" in form:
        try:
            amount = int(form["pay"][0])
        except ValueError: # Bad syntax
            amount = 0
    if amount > 0:
        status = '200 OK'
        response_headers.append(('Content-type', 'application/json'))
        output = json.dumps({"title": "OK, %i credits paid" % amount}, indent=2)
    else:
        status = '402 Payment required'
        response_headers.append(('Content-type', 'application/problem+json'))
        output = json.dumps({"type": "http://errors.bortzmeyer.org/nopay",
                             "title": "You must pay",
                             "detail": "Bitcoin address 1HtNJ6ZFUc9yu9u2qAwB4tGdGwPQasQGax, Ethereum address 0xbe1f2ac71a9703275a4d3ea01a340f378c931740"},
                             indent=2)
    response_headers.append(('Content-Length', str(len(output))))
    start_response(status, response_headers)
    return [output.encode()]

def porn(start_response, environ):
    safe = False # Default value if the user said nothing
    if 'HTTP_PREFER' in list(environ.keys()):
        # Apache/WSGI always give us one Prefer: header even if the client sent several.
        preferences = re.split(r"\s*,\s*", environ['HTTP_PREFER'])
        for pref in preferences:
            if pref.lower() == 'safe':
                safe = True
                break
    begin = """
    <html><head><title>Content</title></head>
    <body>
    <h1>Your content</h1>"""
    end = """
    </body>
    </html>"""
    safe_content = """
    <p>No nudity on this picture: <a title="w:Christopher Voelker [CC BY 3.0 (https://creativecommons.org/licenses/by/3.0)], via Wikimedia Commons" href="https://commons.wikimedia.org/wiki/File:Sphynx_cat_wearing_clothes.jpg"><img width="256" alt="Sphynx cat wearing clothes" src="https://upload.wikimedia.org/wikipedia/commons/thumb/6/60/Sphynx_cat_wearing_clothes.jpg/256px-Sphynx_cat_wearing_clothes.jpg"></a></p>"""
    unsafe_content = """
    <p>Warning, the cat is naked: <a title="Dropdeadpretty [CC BY-SA 4.0 (https://creativecommons.org/licenses/by-sa/4.0)], via Wikimedia Commons" href="https://commons.wikimedia.org/wiki/File:A_Sphynx_cat.jpg"><img width="256" alt="A Sphynx cat" src="https://upload.wikimedia.org/wikipedia/commons/thumb/7/77/A_Sphynx_cat.jpg/256px-A_Sphynx_cat.jpg"></a></p>"""
    if safe:
        output = begin + safe_content + end
    else:
        output = begin + unsafe_content + end
    status = '200 OK'
    response_headers = [('Content-type', 'text/html'),
                        ('Content-Length', str(len(output)))]
    start_response(status, response_headers)
    return [output.encode()]

# draft-ietf-httpapi-deprecation-header and RFC 8594
def limit(start_response, environ):
    output = "Will be deprecated on 2038-01-19T03:14:08Z and removed on 2100-01-01T00:00:00Z.\r\n"
    status = '200 OK'
    response_headers = [('Content-type', 'text/plain'),
                        ('Content-Length', str(len(output))),
                        ('Deprecation', '@2147483648'), # 2038-01-19T03:14:08Z
                        ('Link', '<https://www.bortzmeyer.org/9745.html>; rel="deprecation"; type="text/html"'),
                        ('Sunset', 'Fri, 1 Jan 2100 00:00:00 GMT'),
                        ('Link', '<https://www.bortzmeyer.org/8594.html>; rel="sunset"; type="text/html"')
                        ]
    start_response(status, response_headers)
    return [output.encode()]
    
def default(start_response, path):
    status = '200 OK'
    
    output = """
I'm the default handler, \"%s\" was called.
Are you sure of the URL?\n""" % path
    
    response_headers = [('Content-type', 'text/plain'),
                        ('Content-Length', str(len(output)))]
    start_response(status, response_headers)

    return [output.encode()]

def send_error(start_response, status, output):
    response_headers = [('Content-type', 'text/plain'),
                        ('Content-Length', str(len(output)))]
    start_response(status, response_headers)
    return [output.encode()]

def application(environ, start_response):
    global calls
    global generator
    calls += 1
    ip_client = netaddr.IPAddress(environ['REMOTE_ADDR'])
    if ip_client.version == 4:
            ip_prefix = netaddr.IPNetwork(environ['REMOTE_ADDR'] + "/28")
    elif ip_client.version == 6:
        ip_prefix = netaddr.IPNetwork(environ['REMOTE_ADDR'] + "/64")
    else:
        output = "Unsupported address family \"%s\"\n" % ip_client.version
        send_error(start_response, '400 Unknown IP version', output)
        return [output.encode()]
    if ip_client not in whitelist and ip_prefix not in whitelist:
        if buckets[ip_prefix.cidr].full():
            status = '429 Too many requests'
            output = "%s sent too many requests" % environ['REMOTE_ADDR']
            send_error(start_response, status, output)
            return [output.encode()]
        else:
            buckets[ip_prefix.cidr].add(1)
    if environ['PATH_INFO'] == "/counter":
        return counter(start_response, environ['REMOTE_ADDR'],
                       environ['wsgi.version'])
    elif environ['PATH_INFO'] == "/http":
        return http(start_response, environ)
    elif environ['PATH_INFO'] == "/date-in-cbor":
        return date_in_cbor(start_response, environ)
    elif environ['PATH_INFO'] == "/addresses-in-cbor":
        return addresses_in_cbor(start_response, environ['REMOTE_ADDR'])
    elif environ['PATH_INFO'] == "/payme":
        return payme(start_response, environ)
    elif environ['PATH_INFO'] == "/slow":
        return slow(start_response, environ)
    elif environ['PATH_INFO'] == "/random":
        return randomfile(start_response, environ)
    elif environ['PATH_INFO'] == "/env":
        return env(start_response, environ)
    elif environ['PATH_INFO'] == "/porn":
        return porn(start_response, environ)
    elif environ['PATH_INFO'] == "/limit":
        return limit(start_response, environ)
    else:
        return default(start_response, environ['PATH_INFO'])

if __name__ == '__main__':
    def local_debug(status, headers):
        print(status)
        print(headers)
    # Run the function we want to test
    print(http(local_debug, os.environ))
    if config is not None:
        print("Configuration:")
        print(config)
